#define SMALL_NUMBER	1.e-8
/* 
 *   inverse( original_matrix, inverse_matrix )
 * 
 *    calculate the inverse of a 4x4 matrix
 *
 *     -1     
 *     A  = ___1__ adjoint A
 *         det A
 */


#include <math.h>
#include <stdio.h>
#include "mat.h"

void inverse(view *in, view *out)
{
    int i, j;
    float det;

    /* calculate the adjoint matrix */

    adjoint( in, out );

    /*  calculate the 4x4 determinant
     *  if the determinant is zero, 
     *  then the inverse matrix is not unique.
     */

    det = det4x4( in );

    if ( fabs( det ) < SMALL_NUMBER) {
        printf("Non-singular matrix, no inverse!\n");
        exit(1);
    }

    /* scale the adjoint matrix to get the inverse */

    for (i=0; i<4; i++)
        for(j=0; j<4; j++)
	    (out->rot)[4*i+j] = (out->rot)[4*i+j] / det;
}


/* 
 *   adjoint( original_matrix, inverse_matrix )
 * 
 *     calculate the adjoint of a 4x4 matrix
 *
 *      Let  a   denote the minor determinant of matrix A obtained by
 *           ij
 *
 *      deleting the ith row and jth column from A.
 *
 *                    i+j
 *     Let  b   = (-1)    a
 *          ij            ji
 *
 *    The matrix B = (b  ) is the adjoint of A
 *                     ij
 */

void adjoint(view* in, view* out )
{
    double a1, a2, a3, a4, b1, b2, b3, b4;
    double c1, c2, c3, c4, d1, d2, d3, d4;

    /* assign to individual variable names to aid  */
    /* selecting correct values  */

	a1 = in->rot[0]; b1 = in->rot[1];
	c1 = in->rot[2]; d1 = in->rot[3];

	a2 = in->rot[4]; b2 = in->rot[5];
	c2 = in->rot[6]; d2 = in->rot[7];

	a3 = in->rot[8]; b3 = in->rot[9];
	c3 = in->rot[10]; d3 = in->rot[11];

	a4 = in->rot[12]; b4 = in->rot[13];
	c4 = in->rot[14]; d4 = in->rot[15];


    /* row column labeling reversed since we transpose rows & columns */

    out->rot[0]  =   det3x3( b2, b3, b4, c2, c3, c4, d2, d3, d4);
    out->rot[4]  = - det3x3( a2, a3, a4, c2, c3, c4, d2, d3, d4);
    out->rot[8]  =   det3x3( a2, a3, a4, b2, b3, b4, d2, d3, d4);
    out->rot[12]  = - det3x3( a2, a3, a4, b2, b3, b4, c2, c3, c4);

    out->rot[1]  = - det3x3( b1, b3, b4, c1, c3, c4, d1, d3, d4);
    out->rot[5]  =   det3x3( a1, a3, a4, c1, c3, c4, d1, d3, d4);
    out->rot[9]  = - det3x3( a1, a3, a4, b1, b3, b4, d1, d3, d4);
    out->rot[13]  =   det3x3( a1, a3, a4, b1, b3, b4, c1, c3, c4);

    out->rot[2]  =   det3x3( b1, b2, b4, c1, c2, c4, d1, d2, d4);
    out->rot[6]  = - det3x3( a1, a2, a4, c1, c2, c4, d1, d2, d4);
    out->rot[10]  =   det3x3( a1, a2, a4, b1, b2, b4, d1, d2, d4);
    out->rot[14]  = - det3x3( a1, a2, a4, b1, b2, b4, c1, c2, c4);

    out->rot[3]  = - det3x3( b1, b2, b3, c1, c2, c3, d1, d2, d3);
    out->rot[7]  =   det3x3( a1, a2, a3, c1, c2, c3, d1, d2, d3);
    out->rot[11]  = - det3x3( a1, a2, a3, b1, b2, b3, d1, d2, d3);
    out->rot[15]  =   det3x3( a1, a2, a3, b1, b2, b3, c1, c2, c3);
}
/*
 * double = det4x4( matrix )
 *
 * calculate the determinant of a 4x4 matrix.
 */
double det4x4(view* m )
{
    double ans;
    double a1, a2, a3, a4, b1, b2, b3, b4, c1, c2, c3, c4, d1, d2, d3, 			d4;


    /* assign to individual variable names to aid selecting */
	/*  correct rots */

	a1 = m->rot[0]; b1 = m->rot[1];
	c1 = m->rot[2]; d1 = m->rot[3];

	a2 = m->rot[4]; b2 = m->rot[5];
	c2 = m->rot[6]; d2 = m->rot[7];

	a3 = m->rot[8]; b3 = m->rot[9];
	c3 = m->rot[10]; d3 = m->rot[11];

	a4 = m->rot[12]; b4 = m->rot[13];
	c4 = m->rot[14]; d4 = m->rot[15];

    ans = a1 * det3x3( b2, b3, b4, c2, c3, c4, d2, d3, d4)
        - b1 * det3x3( a2, a3, a4, c2, c3, c4, d2, d3, d4)
        + c1 * det3x3( a2, a3, a4, b2, b3, b4, d2, d3, d4)
        - d1 * det3x3( a2, a3, a4, b2, b3, b4, c2, c3, c4);
    return ans;
}



/*
 * double = det3x3(  a1, a2, a3, b1, b2, b3, c1, c2, c3 )
 * 
 * calculate the determinant of a 3x3 matrix
 * in the form
 *
 *     | a1,  b1,  c1 |
 *     | a2,  b2,  c2 |
 *     | a3,  b3,  c3 |
 */

double det3x3(float a1,float a2,float a3,float b1,float b2,float b3,float c1,float c2,float c3 )
{
    double ans;

    ans = a1 * det2x2( b2, b3, c2, c3 )
        - b1 * det2x2( a2, a3, c2, c3 )
        + c1 * det2x2( a2, a3, b2, b3 );
    return ans;
}

/*
 * double = det2x2( double a, double b, double c, double d )
 * 
 * calculate the determinant of a 2x2 matrix.
 */

double det2x2(float a,float b,float c,float d)
{
    double ans;
    ans = a * d - b * c;
    return ans;
}





